/// <reference types="node" />
import { ILogger, IPargv } from 'pargv';
import { EventEmitter } from 'events';
import { IMustr, IMustrOptions, ITemplates, ITemplate, BeforeRender, IMetadata, InjectCallback, NodeCallback } from './interfaces';
export declare class Mustr extends EventEmitter implements IMustr {
    cwd: string;
    pargv: IPargv;
    log: ILogger;
    registerPath: string;
    templatesPath: string;
    outputPath: string;
    loaded: boolean;
    templatesGlob: string[];
    templates: ITemplates;
    options: IMustrOptions;
    constructor(options?: IMustrOptions);
    /**
     * Has Matter
     * Tests if string contains front matter.
     *
     * @param str the string to inspect.
     */
    private hasMatter(str);
    /**
     * Has Template
     * Loosly check if string contains template chars.
     *
     * @param str the string to inspect.
     */
    private hasTemplate(str);
    /**
     * Try Require
     * Tries to require a module logging error if failed.
     *
     * @param path the module path to be required.
     */
    private tryRequire(path);
    /**
     * Normalize Extension
     * Ensures extension begins with "."
     *
     * @param ext the extension to be normalized.
     */
    private normalizedExt(ext);
    isValid(): void;
    /**
     * Help
     * Displays help and usage information.
     */
    help(): void;
    /**
     * Init
     * Initializes Mustr for current project
     *
     * @param force when true forces init.
     */
    init(force?: boolean): void;
    /**
     * Load
     * Loads the template paths and config.
     */
    load(): IMustr;
    /**
     * Normalize
     * Parses/normalizes a template name, path or static template string.
     * When passing a name it searches loaded template paths and matches
     * the filename to the template name passed.
     *
     * @param template the template name, path or static template string.
     */
    normalize(template: string): ITemplate;
    /**
     * Register
     * Registers a template with Mustr.
     *
     * @param template the template name, path or static string to be registered.
     * @param metadata optional metadata to be merged with front matter.
     * @param action the action to be called on generate.
     */
    register(name: string, template: string, partials?: string | string[] | BeforeRender, beforeRender?: BeforeRender): IMustr;
    /**
     * Generate
     * Generates and renders a template.
     *
     * @param name the name of the template to generate.
     * @param output the output name/path to use when rendering.
     * @param options any metadata to merge for rendering.
     */
    generate(name?: string, output?: string | IMetadata, options?: IMetadata): ITemplate;
    /**
     * Render
     * Renders the template and writes out result.
     *
     * @param path the path to render the template to.
     * @param body the body or template string to be used in compiling.
     * @param data the data to use when compiling body for render.
     */
    render(path: string | ITemplate, body?: string, data?: {
        [key: string]: any;
    }, partials?: {
        [key: string]: any;
    }): void;
    /**
     * Inject
     * Creates a file stream reading chunks until match.
     * Upon match the provided value is injected.
     * NOTE: lines are auto updated with the correct
     * line ending for returns.
     *
     * @param filename the file name to inject into.
     * @param search the string or expression to match.
     * @param value the value or callback for injecting upon match.
     * @param replace when true the matched value is replaced.
     * @param done the callback on file written.
     */
    inject(filename: string, search: string | RegExp, value: string | string[] | InjectCallback, replace: boolean | NodeCallback, done?: NodeCallback): void;
    /**
     * Transform To
     * Transforms a string to the desired casing.
     *
     * @param str the string to be transformed.
     * @param to the casing to be transformed to using lodash.
     */
    transformCase(str: string, to: string): string;
}
