"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var fs_extra_1 = require("fs-extra");
var readline = require("readline");
var path_1 = require("path");
var pargv_1 = require("pargv");
var os_1 = require("os");
var glob = require("glob");
var lodash_1 = require("lodash");
var events_1 = require("events");
var fm = require("front-matter");
var Mustache = require("mustache");
var defaults = {
    configDir: './mustr',
    outputDir: './src',
    templateExt: '.tpl'
};
var chalk = new pargv_1.Chalk();
var Mustr = (function (_super) {
    __extends(Mustr, _super);
    function Mustr(options) {
        var _this = _super.call(this) || this;
        // Utils
        _this.cwd = process.cwd();
        // Parsed templates.
        _this.templates = {};
        _this.options = lodash_1.extend({}, defaults, options);
        _this.log = pargv_1.logger();
        _this.pargv = new pargv_1.Pargv();
        _this.templatesPath = path_1.resolve(_this.cwd, path_1.join(_this.options.configDir, '/**/*.tpl'));
        // Resolve the base directory.
        _this.outputPath = path_1.resolve(_this.cwd, _this.options.outputDir);
        // Resolve the config path.
        _this.registerPath = path_1.resolve(_this.cwd, _this.options.configDir, 'register.js');
        // Normalized the template extension.
        _this.options.templateExt = _this.normalizedExt(_this.options.templateExt);
        return _this;
    }
    /**
     * Has Matter
     * Tests if string contains front matter.
     *
     * @param str the string to inspect.
     */
    Mustr.prototype.hasMatter = function (str) {
        return str.indexOf('---') > -1;
    };
    /**
     * Has Template
     * Loosly check if string contains template chars.
     *
     * @param str the string to inspect.
     */
    Mustr.prototype.hasTemplate = function (str) {
        return str.indexOf('{{') > -1;
    };
    /**
     * Try Require
     * Tries to require a module logging error if failed.
     *
     * @param path the module path to be required.
     */
    Mustr.prototype.tryRequire = function (path) {
        try {
            return require(path);
        }
        catch (ex) {
            this.log.error("failed to require module at " + path + ".").exit();
        }
    };
    /**
     * Normalize Extension
     * Ensures extension begins with "."
     *
     * @param ext the extension to be normalized.
     */
    Mustr.prototype.normalizedExt = function (ext) {
        return "." + ext.replace(/^\./, '');
    };
    Mustr.prototype.isValid = function () {
    };
    /**
     * Help
     * Displays help and usage information.
     */
    Mustr.prototype.help = function () {
        var padBtm = [0, 0, 1, 0];
        var msg = '- See additional required/optional arguments for each command below. -\n';
        this.pargv
            .logo('Mustr', 'cyan')
            .ui(95)
            .join(chalk.magenta('Usage:'), 'mu', chalk.cyan('<cmd>'), '\n')
            .div({ text: chalk.bgBlue.white(msg) })
            .div({ text: chalk.cyan('help, h'), width: 35, padding: padBtm }, { text: chalk.gray('displays help and usage information.'), width: 40, padding: padBtm })
            .div({ text: chalk.cyan('init, i'), width: 35, padding: padBtm }, { text: chalk.gray('initialize the application for use with Mustr.'), width: 40, padding: padBtm })
            .div({ text: chalk.cyan('generate, g') + " " + chalk.white('<template>') + " " + chalk.magenta('<name>'), width: 35, padding: padBtm }, { text: chalk.gray('generates and renders a template.'), width: 40, padding: padBtm })
            .div({ text: chalk.white('<template>'), width: 35, padding: [0, 2, 0, 2] }, { text: chalk.gray('template to generate and compile.'), width: 40 }, { text: chalk.red('[required]'), align: 'right' })
            .div({ text: chalk.white('<name>'), width: 35, padding: [0, 2, 0, 2] }, { text: chalk.gray('output name for template'), width: 40 }, { text: chalk.red('[required]'), align: 'right' })
            .show();
    };
    /**
     * Init
     * Initializes Mustr for current project
     *
     * @param force when true forces init.
     */
    Mustr.prototype.init = function (force) {
        var configPath = path_1.resolve(this.cwd, 'mustr.json');
        var blueprintsPath = path_1.resolve(this.cwd, 'mustr');
        if (this.pargv.f)
            force = true;
        // Ensure not already initialized.
        if (!force && (fs_extra_1.existsSync(configPath) || fs_extra_1.existsSync(blueprintsPath)))
            this.log.warn('project already initialized, use mu.init(true) or mu init -f from cli to force.').exit();
        // Write out the mustr.json config.
        fs_extra_1.writeFileSync(configPath, JSON.stringify(defaults, null, 2));
        // Copy the blueprints.
        fs_extra_1.copySync(path_1.resolve(__dirname, 'blueprints'), blueprintsPath);
        this.log.info('successfully initialized Mustr.').exit();
    };
    /**
     * Load
     * Loads the template paths and config.
     */
    Mustr.prototype.load = function () {
        if (!fs_extra_1.existsSync(this.registerPath))
            this.log.error("failed to resolve register configuration at " + this.registerPath + ".").exit();
        // Load template paths..
        this.templatesGlob = glob.sync(this.templatesPath);
        // Require the config
        var config = this.tryRequire(this.registerPath);
        if (!lodash_1.isFunction(config))
            this.log
                .error("failed to load config expected function but got type of " + typeof config + ".")
                .exit();
        // Call/load the configuration.
        config(this);
        this.loaded = true;
        return this;
    };
    /**
     * Normalize
     * Parses/normalizes a template name, path or static template string.
     * When passing a name it searches loaded template paths and matches
     * the filename to the template name passed.
     *
     * @param template the template name, path or static template string.
     */
    Mustr.prototype.normalize = function (template) {
        var rawTemplate, templatePath, templatePathNormalized, templateName, templateExt;
        var EXT_EXP = new RegExp(this.options.templateExt + "$");
        var parsedPath, parsedNormalized;
        // When has extension readfile.
        if (EXT_EXP.test(template)) {
            // If template doesn't exist error & exit.
            if (!fs_extra_1.existsSync(template))
                this.log.error("failed to resolve template at path " + template + ".").exit();
            // Save the template path.
            templatePath = template;
        }
        else if (!this.hasMatter(template) && !this.hasTemplate(template)) {
            var filtered = this.templatesGlob.filter(function (t) {
                return path_1.parse(t).name === template;
            })[0];
            if (!filtered) {
                this.log
                    .error("failed to resolve template for name " + template + ".")
                    .exit();
            }
            templatePath = filtered;
        }
        // If template path exists normalize it.
        if (templatePath) {
            // We'll resolve the template later
            // set undefined as this is NOT a
            // static template.
            template = undefined;
            // Remove the template ext.
            templatePathNormalized = templatePath.replace(EXT_EXP, '');
            // Parse paths.
            parsedPath = path_1.parse(templatePath);
            parsedNormalized = path_1.parse(templatePathNormalized);
            // Get the name from the path.
            templateName = parsedPath.name;
            // Template may contain extension in path.
            // So we make sure we used the normalized path
            // which exclusdes the template extension.
            templateExt = parsedNormalized.ext;
        }
        else {
            rawTemplate = template;
        }
        // Get the template's configuration.
        var config = {
            name: templateName,
            path: templatePath,
            ext: templateExt,
            pathNormalized: templatePathNormalized,
            raw: rawTemplate,
            partials: {}
        };
        // Return ITemplate
        return config;
    };
    /**
     * Register
     * Registers a template with Mustr.
     *
     * @param template the template name, path or static string to be registered.
     * @param metadata optional metadata to be merged with front matter.
     * @param action the action to be called on generate.
     */
    Mustr.prototype.register = function (name, template, partials, beforeRender) {
        var _this = this;
        if (lodash_1.isFunction(partials)) {
            beforeRender = partials;
            partials = undefined;
        }
        if (lodash_1.isString(partials))
            partials = [partials];
        var _partials = partials || [];
        // Parse the supplied template.
        var parsed = this.normalize(template);
        // Iterate partials and load.
        _partials.forEach(function (p) {
            // Check if already loaded.
            if (_this.templates[p]) {
                parsed.partials[p] = _this.templates[p];
            }
            else {
                var parsedPartial = _this.normalize(p);
                parsed.partials[parsedPartial.name] = parsedPartial;
                // Add the partial to the templates collection.
                _this.templates[parsedPartial.name] = parsedPartial;
            }
        });
        if (beforeRender)
            parsed.beforeRender = beforeRender;
        // Register the template with the template collection.
        this.templates[name] = parsed;
        return this;
    };
    /**
     * Generate
     * Generates and renders a template.
     *
     * @param name the name of the template to generate.
     * @param output the output name/path to use when rendering.
     * @param options any metadata to merge for rendering.
     */
    Mustr.prototype.generate = function (name, output, options) {
        var rendered, ext, parsedOutput, outputName, outputPath;
        // Enable options as second arg.
        if (lodash_1.isPlainObject(output)) {
            options = output;
            output = undefined;
        }
        // set output to name if undefined.
        if (!lodash_1.isUndefined(name) && lodash_1.isUndefined(output))
            output = name;
        name = name || this.pargv.getCmd(0);
        output = output || this.pargv.getCmd(1);
        options = options || this.pargv.getFlags();
        // Ensure user templates/config loaded.
        if (!this.loaded)
            this.load();
        // Ensure we have a template and an output name.
        if (lodash_1.isUndefined(name) || lodash_1.isUndefined(output))
            this.log.error('cannot generate template using template name or output name of undefined.').exit();
        var template = this.templates[name];
        // Ensure the template exists.
        if (!template)
            this.log.error("the template " + name + " could not be found.").exit();
        // If contains path then not static defined
        // we need to read the template file.
        if (template.path)
            template.raw = fs_extra_1.readFileSync(template.path).toString();
        // We need to inspect partials.
        // Since its a partial we just need to
        // read the file.
        var partialKeys = Object.keys(template.partials);
        partialKeys.forEach(function (p) {
            var partial = template.partials[p];
            if (partial.path)
                partial.raw = fs_extra_1.readFileSync(partial.path).toString();
            // Update the partial template.
            template.partials[p] = partial;
        });
        // If we don't have template.raw string to be parsed
        // we can't continue.
        if (lodash_1.isUndefined(template.raw))
            this.log.error('cannot generate template with raw template string of undefined.').exit();
        // Parse the template for body and front matter attributes.
        var matter = fm(template.raw);
        var attrs = matter.attributes;
        var tmpConfig = attrs.config || {};
        delete attrs.config;
        template.metadata = attrs;
        template.body = matter.body;
        // Merge in the parsed config with template config.
        template = lodash_1.extend({}, tmpConfig, template);
        // Parse the output path/name.
        parsedOutput = path_1.parse(output);
        // Get the output name.
        outputName = parsedOutput.name;
        // Ensure ext from output or defined in template config.
        ext = parsedOutput.ext || tmpConfig.ext;
        // Set the component name in metadata.
        template.metadata.component = outputName;
        // Check component casing and suffix.
        if (template.suffix) {
            template.metadata.component += " " + template.suffix;
            outputName += "." + template.suffix;
        }
        // Ensure full template output path.
        outputPath = path_1.resolve(this.outputPath, parsedOutput.dir || '', outputName + ext);
        outputPath = outputPath.toLowerCase();
        // Update the template object with
        // the output name and output path.
        template.outputPath = outputPath;
        // Check if casing is specified.
        if (template.casing || template.suffix) {
            var casing = template.casing ? template.casing : 'title';
            template.metadata.component =
                this.transformCase(template.metadata.component, casing);
        }
        // Merge parsed options with template metadata.
        template.metadata = lodash_1.extend({}, template.metadata, options);
        return template;
    };
    /**
     * Render
     * Renders the template and writes out result.
     *
     * @param path the path to render the template to.
     * @param body the body or template string to be used in compiling.
     * @param data the data to use when compiling body for render.
     */
    Mustr.prototype.render = function (path, body, data, partials) {
        if (lodash_1.isPlainObject(path)) {
            var template_1 = path;
            data = template_1.metadata;
            body = template_1.body;
            path = template_1.outputPath;
            partials = {};
            // If get the raw strings.
            var partialKeys = Object.keys(template_1.partials);
            partialKeys.forEach(function (p) {
                var partial = template_1.partials[p];
                partials[p] = partial.raw;
            });
        }
        partials = partials || {};
        // Parse the output path.
        var parsedPath = path_1.parse(path);
        // Ensure the directory structure exists.
        fs_extra_1.ensureDirSync(parsedPath.dir);
        // Write out the rendered template.
        fs_extra_1.writeFileSync(path, Mustache.render(body, data, partials));
    };
    /**
     * Inject
     * Creates a file stream reading chunks until match.
     * Upon match the provided value is injected.
     * NOTE: lines are auto updated with the correct
     * line ending for returns.
     *
     * @param filename the file name to inject into.
     * @param search the string or expression to match.
     * @param value the value or callback for injecting upon match.
     * @param replace when true the matched value is replaced.
     * @param done the callback on file written.
     */
    Mustr.prototype.inject = function (filename, search, value, replace, done) {
        var data = '';
        var exp = search;
        if (lodash_1.isFunction(replace)) {
            done = replace;
            replace = undefined;
        }
        if (!fs_extra_1.existsSync(filename))
            this.log.error("failed to inject the file " + filename + " cannot be found.").exit();
        // If is string convert to expression.
        if (lodash_1.isString(exp))
            exp = new RegExp(exp);
        var rl = readline.createInterface({
            input: fs_extra_1.createReadStream(filename)
        });
        rl.on('line', function (line) {
            if (lodash_1.isFunction(value)) {
                // Pause the stream.
                rl.pause();
                // Pass the current chunk and the pending data.
                value(line, data, function (_line, _data) {
                    // If line just push.
                    if (_line)
                        data += (_line + os_1.EOL);
                    else if (_data)
                        data = (_data + os_1.EOL);
                    else
                        data += (line + os_1.EOL);
                    // Resume the stream.
                    rl.resume();
                });
            }
            else {
                // If matches expression inject.
                if (exp.test(line)) {
                    // Check if value is array.
                    if (Array.isArray(value))
                        value = value.join(os_1.EOL);
                    else
                        value += os_1.EOL;
                    if (!replace)
                        data += (line + os_1.EOL);
                    data += value;
                }
                else {
                    data += (line + os_1.EOL);
                }
            }
        });
        rl.on('close', function () {
            fs_extra_1.writeFile(filename, data, done);
        });
    };
    /**
     * Transform To
     * Transforms a string to the desired casing.
     *
     * @param str the string to be transformed.
     * @param to the casing to be transformed to using lodash.
     */
    Mustr.prototype.transformCase = function (str, to) {
        var caseMap = {
            lower: lodash_1.lowerCase,
            upper: lodash_1.upperCase,
            capitalize: lodash_1.capitalize,
            camel: lodash_1.camelCase,
            lowerCase: lodash_1.lowerCase,
            upperCase: lodash_1.upperCase,
            camelCase: lodash_1.camelCase,
            title: function (str) {
                return lodash_1.startCase(lodash_1.toLower(str));
            }
        };
        caseMap.titleCase = caseMap.title;
        if (!caseMap[to])
            this.log.error("invalid transform case " + to + ".").exit();
        return caseMap[to](str).replace(/\s/g, '');
    };
    return Mustr;
}(events_1.EventEmitter));
exports.Mustr = Mustr;
//# sourceMappingURL=index.js.map