
import { readFile, readFileSync, writeFileSync, writeFile, existsSync, copySync, ensureDirSync, createReadStream } from 'fs-extra';
import * as readline from 'readline';
import { join, resolve, parse, ParsedPath, relative, sep } from 'path';
import { Pargv, logger, ILogger, IPargv, Chalk } from 'pargv';
import { EOL } from 'os';
import * as glob from 'glob';
import {
    extend as _extend, get as _get,
    set as _set, unset as _unset,
    isUndefined as _isUndefined,
    isPlainObject as _isPlainObject,
    isFunction as _isFunction,
    isString as _isString,
    camelCase as _camelCase,
    lowerCase as _lowerCase,
    upperCase as _upperCase,
    startCase as _startCase,
    capitalize as _capitalize,
    toLower as _toLower
} from 'lodash';

import { EventEmitter } from 'events';
import * as fm from 'front-matter';
import * as Mustache from 'mustache';
import { IMustr, IMustrOptions, ITemplates, ITemplate, BeforeRender, IMetadata, InjectCallback, NodeCallback } from './interfaces';

const defaults: IMustrOptions = {
    configDir: './mustr',
    outputDir: './src',
    templateExt: '.tpl'
};

const chalk = new Chalk();

export class Mustr extends EventEmitter implements IMustr {

    // Utils
    cwd: string = process.cwd();
    pargv: IPargv;
    log: ILogger;

    // Paths
    registerPath: string;
    templatesPath: string;
    outputPath: string;
    loaded: boolean;

    // Array of template paths.
    templatesGlob: string[];

    // Parsed templates.
    templates: ITemplates = {};

    // Options
    options: IMustrOptions;

    constructor(options?: IMustrOptions) {

        super();

        this.options = _extend({}, defaults, options);

        this.log = logger();
        this.pargv = new Pargv();

        this.templatesPath = resolve(this.cwd, join(this.options.configDir, '/**/*.tpl'));

        // Resolve the base directory.
        this.outputPath = resolve(this.cwd, this.options.outputDir);

        // Resolve the config path.
        this.registerPath = resolve(this.cwd, this.options.configDir, 'register.js');

        // Normalized the template extension.
        this.options.templateExt = this.normalizedExt(this.options.templateExt);

    }

    /**
     * Has Matter
     * Tests if string contains front matter.
     *
     * @param str the string to inspect.
     */
    private hasMatter(str) {
        return str.indexOf('---') > -1;
    }

    /**
     * Has Template
     * Loosly check if string contains template chars.
     *
     * @param str the string to inspect.
     */
    private hasTemplate(str) {
        return str.indexOf('{{') > -1;
    }

    /**
     * Try Require
     * Tries to require a module logging error if failed.
     *
     * @param path the module path to be required.
     */
    private tryRequire(path: string) {
        try {
            return require(path);
        }
        catch (ex) {
            this.log.error(`failed to require module at ${path}.`).exit();
        }
    }

    /**
     * Normalize Extension
     * Ensures extension begins with "."
     *
     * @param ext the extension to be normalized.
     */
    private normalizedExt(ext): string {
        return `.${ext.replace(/^\./, '')}`;
    }

    isValid() {

    }

    /**
     * Help
     * Displays help and usage information.
     */
    help(): void {

        const padBtm = [0, 0, 1, 0];

        const msg =
            '- See additional required/optional arguments for each command below. -\n';


        this.pargv
            .logo('Mustr', 'cyan')
            .ui(95)
            // .join(chalk.magenta('Usage:'), 'mu', chalk.cyan('<cmd>'), '<type> [options]\n')
            .join(chalk.magenta('Usage:'), 'mu', chalk.cyan('<cmd>'), '\n')
            .div({ text: chalk.bgBlue.white(msg) })
            .div(
            { text: chalk.cyan('help, h'), width: 35, padding: padBtm },
            { text: chalk.gray('displays help and usage information.'), width: 40, padding: padBtm }
            )
            .div(
            { text: chalk.cyan('init, i'), width: 35, padding: padBtm },
            { text: chalk.gray('initialize the application for use with Mustr.'), width: 40, padding: padBtm }
            )
            .div(
            { text: `${chalk.cyan('generate, g')} ${chalk.white('<template>')} ${chalk.magenta('<name>')}`, width: 35, padding: padBtm },
            { text: chalk.gray('generates and renders a template.'), width: 40, padding: padBtm }
            )
            .div(
            { text: chalk.white('<template>'), width: 35, padding: [0, 2, 0, 2] },
            { text: chalk.gray('template to generate and compile.'), width: 40 },
            { text: chalk.red('[required]'), align: 'right' }
            )
            .div(
            { text: chalk.white('<name>'), width: 35, padding: [0, 2, 0, 2] },
            { text: chalk.gray('output name for template'), width: 40 },
            { text: chalk.red('[required]'), align: 'right' }
            )
            // .div(
            // { text: chalk.white('--data, -d'), width: 35, padding: [0, 2, 0, 2] },
            // { text: chalk.gray('inline data for compile.'), width: 40 },
            // { text: chalk.gray('[optional]'), align: 'right' }
            // )
            .show();

    }

    /**
     * Init
     * Initializes Mustr for current project
     *
     * @param force when true forces init.
     */
    init(force?: boolean): void {

        const configPath = resolve(this.cwd, 'mustr.json');
        const blueprintsPath = resolve(this.cwd, 'mustr');

        if (this.pargv.f)
            force = true;

        // Ensure not already initialized.
        if (!force && (existsSync(configPath) || existsSync(blueprintsPath)))
            this.log.warn('project already initialized, use mu.init(true) or mu init -f from cli to force.').exit();

        // Write out the mustr.json config.
        writeFileSync(configPath, JSON.stringify(defaults, null, 2));

        // Copy the blueprints.
        copySync(resolve(__dirname, 'blueprints'), blueprintsPath);

        this.log.info('successfully initialized Mustr.').exit();

    }

    /**
     * Load
     * Loads the template paths and config.
     */
    load(): IMustr {

        if (!existsSync(this.registerPath))
            this.log.error(`failed to resolve register configuration at ${this.registerPath}.`).exit();

        // Load template paths..
        this.templatesGlob = glob.sync(this.templatesPath);

        // Require the config
        const config = this.tryRequire(this.registerPath);

        if (!_isFunction(config))
            this.log
                .error(`failed to load config expected function but got type of ${typeof config}.`)
                .exit();

        // Call/load the configuration.
        config(this);

        this.loaded = true;

        return this;

    }

    /**
     * Normalize
     * Parses/normalizes a template name, path or static template string.
     * When passing a name it searches loaded template paths and matches
     * the filename to the template name passed.
     *
     * @param template the template name, path or static template string.
     */
    normalize(template: string): ITemplate {

        let rawTemplate, templatePath, templatePathNormalized,
            templateName, templateExt;
        const EXT_EXP = new RegExp(`${this.options.templateExt}$`);

        let parsedPath, parsedNormalized;

        // When has extension readfile.
        if (EXT_EXP.test(template)) {

            // If template doesn't exist error & exit.
            if (!existsSync(template))
                this.log.error(`failed to resolve template at path ${template}.`).exit();

            // Save the template path.
            templatePath = template;

        }

        // If not a static template try to
        // lookup from loaded template paths and read.
        else if (!this.hasMatter(template) && !this.hasTemplate(template)) {

            const filtered = this.templatesGlob.filter((t) => {
                return parse(t).name === template;
            })[0];

            if (!filtered) {

                this.log
                    .error(`failed to resolve template for name ${template}.`)
                    .exit();
            }

            templatePath = filtered;

        }

        // If template path exists normalize it.
        if (templatePath) {

            // We'll resolve the template later
            // set undefined as this is NOT a
            // static template.
            template = undefined;

            // Remove the template ext.
            templatePathNormalized = templatePath.replace(EXT_EXP, '');

            // Parse paths.
            parsedPath = parse(templatePath);
            parsedNormalized = parse(templatePathNormalized);

            // Get the name from the path.
            templateName = parsedPath.name;

            // Template may contain extension in path.
            // So we make sure we used the normalized path
            // which exclusdes the template extension.
            templateExt = parsedNormalized.ext;

        }

        // Otherwise this is a static template
        // set raw to be parsed later.
        else {
            rawTemplate = template;
        }

        // Get the template's configuration.
        const config: ITemplate = {
            name: templateName,
            path: templatePath,
            ext: templateExt,
            pathNormalized: templatePathNormalized,
            raw: rawTemplate,
            partials: {}
        };

        // Return ITemplate
        return config;

    }

    /**
     * Register
     * Registers a template with Mustr.
     *
     * @param template the template name, path or static string to be registered.
     * @param metadata optional metadata to be merged with front matter.
     * @param action the action to be called on generate.
     */
    register(name: string, template: string, partials?: string | string[] | BeforeRender, beforeRender?: BeforeRender): IMustr {

        if (_isFunction(partials)) {
            beforeRender = partials;
            partials = undefined;
        }

        if (_isString(partials))
            partials = [partials];

        const _partials: string[] = partials as string[] || [];

        // Parse the supplied template.
        const parsed = this.normalize(template);

        // Iterate partials and load.
        _partials.forEach((p) => {

            // Check if already loaded.
            if (this.templates[p]) {
                parsed.partials[p] = this.templates[p];
            }

            // otherwise normalize and try to
            // load partial automatically.
            else {
                const parsedPartial = this.normalize(p);
                parsed.partials[parsedPartial.name] = parsedPartial;
                // Add the partial to the templates collection.
                this.templates[parsedPartial.name] = parsedPartial;
            }

        });

        if (beforeRender)
            parsed.beforeRender = beforeRender;

        // Register the template with the template collection.
        this.templates[name] = parsed;

        return this;

    }

    /**
     * Generate
     * Generates and renders a template.
     *
     * @param name the name of the template to generate.
     * @param output the output name/path to use when rendering.
     * @param options any metadata to merge for rendering.
     */
    generate(name?: string, output?: string | IMetadata, options?: IMetadata): ITemplate {

        let rendered, ext, parsedOutput, outputName, outputPath;

        // Enable options as second arg.
        if (_isPlainObject(output)) {
            options = output as IMetadata;
            output = undefined;
        }

        // set output to name if undefined.
        if (!_isUndefined(name) && _isUndefined(output))
            output = name;

        name = name || this.pargv.getCmd(0);
        output = output || this.pargv.getCmd(1);
        options = options || this.pargv.getFlags();

        // Ensure user templates/config loaded.
        if (!this.loaded)
            this.load();

        // Ensure we have a template and an output name.
        if (_isUndefined(name) || _isUndefined(output))
            this.log.error('cannot generate template using template name or output name of undefined.').exit();

        let template = this.templates[name];

        // Ensure the template exists.
        if (!template)
            this.log.error(`the template ${name} could not be found.`).exit();

        // If contains path then not static defined
        // we need to read the template file.
        if (template.path)
            template.raw = readFileSync(template.path).toString();

        // We need to inspect partials.
        // Since its a partial we just need to
        // read the file.
        const partialKeys = Object.keys(template.partials);
        partialKeys.forEach((p) => {

            const partial: ITemplate = template.partials[p];

            if (partial.path)
                partial.raw = readFileSync(partial.path).toString();

            // Update the partial template.
            template.partials[p] = partial;

        });

        // If we don't have template.raw string to be parsed
        // we can't continue.
        if (_isUndefined(template.raw))
            this.log.error('cannot generate template with raw template string of undefined.').exit();

        // Parse the template for body and front matter attributes.
        const matter = fm(template.raw);
        const attrs = matter.attributes;
        const tmpConfig = attrs.config || {};
        delete attrs.config;

        template.metadata = attrs;
        template.body = matter.body;

        // Merge in the parsed config with template config.
        template = _extend({}, tmpConfig, template);

        // Parse the output path/name.
        parsedOutput = parse(output as string);

        // Get the output name.
        outputName = parsedOutput.name;

        // Ensure ext from output or defined in template config.
        ext = parsedOutput.ext || tmpConfig.ext;

        // Set the component name in metadata.
        template.metadata.component = outputName;

        // Check component casing and suffix.
        if (template.suffix) {
            template.metadata.component += ` ${template.suffix}`;
            outputName += `.${template.suffix}`;
        }

        // Ensure full template output path.
        outputPath = resolve(this.outputPath, parsedOutput.dir || '', outputName + ext);
        outputPath = outputPath.toLowerCase();

        // Update the template object with
        // the output name and output path.
        template.outputPath = outputPath;

        // Check if casing is specified.
        if (template.casing || template.suffix) {
            const casing = template.casing ? template.casing : 'title';
            template.metadata.component =
                this.transformCase(template.metadata.component, casing);
        }

        // Merge parsed options with template metadata.
        template.metadata = _extend({}, template.metadata, options);

        return template;

    }

    /**
     * Render
     * Renders the template and writes out result.
     *
     * @param path the path to render the template to.
     * @param body the body or template string to be used in compiling.
     * @param data the data to use when compiling body for render.
     */
    render(path: string | ITemplate, body?: string, data?: { [key: string]: any }, partials?: { [key: string]: any }): void {

        if (_isPlainObject(path)) {

            const template: ITemplate = path as ITemplate;
            data = template.metadata;
            body = template.body;
            path = template.outputPath;

            partials = {};

            // If get the raw strings.
            const partialKeys = Object.keys(template.partials);
            partialKeys.forEach((p) => {
                const partial: ITemplate = template.partials[p];
                partials[p] = partial.raw;
            });

        }

        partials = partials || {};

        // Parse the output path.
        const parsedPath = parse(path as string);

        // Ensure the directory structure exists.
        ensureDirSync(parsedPath.dir);

        // Write out the rendered template.
        writeFileSync(path as string, Mustache.render(body, data, partials));

    }

    /**
     * Inject
     * Creates a file stream reading chunks until match.
     * Upon match the provided value is injected.
     * NOTE: lines are auto updated with the correct
     * line ending for returns.
     *
     * @param filename the file name to inject into.
     * @param search the string or expression to match.
     * @param value the value or callback for injecting upon match.
     * @param replace when true the matched value is replaced.
     * @param done the callback on file written.
     */
    inject(filename: string, search: string | RegExp, value: string | string[] | InjectCallback, replace: boolean | NodeCallback, done?: NodeCallback): void {

        let data = '';
        let exp: RegExp = search as RegExp;

        if (_isFunction(replace)) {
            done = replace;
            replace = undefined;
        }

        if (!existsSync(filename))
            this.log.error(`failed to inject the file ${filename} cannot be found.`).exit();

        // If is string convert to expression.
        if (_isString(exp))
            exp = new RegExp(exp);

        const rl = readline.createInterface({
            input: createReadStream(filename)
        });

        rl.on('line', (line) => {

            if (_isFunction(value)) {

                // Pause the stream.
                rl.pause();

                // Pass the current chunk and the pending data.
                value(line, data, (_line, _data) => {

                    // If line just push.
                    if (_line)
                        data += (_line + EOL);

                    // if pending data passed
                    // update the entire data string.
                    else if (_data)
                        data = (_data + EOL);

                    // Otherwise just update using existing line.
                    else
                        data += (line + EOL);

                    // Resume the stream.
                    rl.resume();

                });

            }

            else {

                // If matches expression inject.

                if (exp.test(line)) {

                    // Check if value is array.
                    if (Array.isArray(value))
                        value = value.join(EOL);
                    else
                        value += EOL;

                    if (!replace)
                        data += (line + EOL);

                    data += value;

                }

                else {

                    data += (line + EOL);

                }

            }

        });

        rl.on('close', () => {
            writeFile(filename, data, done);
        });

    }

    /**
     * Transform To
     * Transforms a string to the desired casing.
     *
     * @param str the string to be transformed.
     * @param to the casing to be transformed to using lodash.
     */
    transformCase(str: string, to: string): string {

        let caseMap: any = {
            lower: _lowerCase,
            upper: _upperCase,
            capitalize: _capitalize,
            camel: _camelCase,
            lowerCase: _lowerCase,
            upperCase: _upperCase,
            camelCase: _camelCase,
            title: (str) => {
                return _startCase(_toLower(str));
            }
        };

        caseMap.titleCase = caseMap.title;

        if (!caseMap[to])
            this.log.error(`invalid transform case ${to}.`).exit();

        return caseMap[to](str).replace(/\s/g, '');

    }


}